// Copyright 2021 The Casbin Authors. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { SchemaDirectiveVisitor } from '@graphql-tools/utils';
import { defaultFieldResolver, DirectiveLocation, GraphQLDirective, GraphQLNonNull, GraphQLScalarType, GraphQLString, } from 'graphql';
import { buildPathCache, CasbinAuthTypeName, CasbinContextEnforcerKey, CasbinContextPathCacheKey, CasbinContextRoleKey, Info2Path, PathAndAction2Key, } from './utils';
import { EnforcerNotFoundError, PermissionInsufficientError } from './error';
/**
 * Apply directive to paticular field.
 */
export class CanDirective extends SchemaDirectiveVisitor {
    /**
     * internal function, be called to get directive info.
     * You should also implicit define directive in the GraphQL SDL after graphql-js 14.0.0
     * @param directiveName
     * @param schema
     */
    static getDirectiveDeclaration(directiveName, schema) {
        return new GraphQLDirective({
            name: 'can',
            locations: [DirectiveLocation.FIELD_DEFINITION],
            args: {
                who: {
                    type: GraphQLString,
                    defaultValue: '*',
                },
            },
        });
    }
    /**
     * internal function, be called to visit fields.
     * @param field
     */
    visitFieldDefinition(field) {
        const { resolve = defaultFieldResolver } = field;
        const { who } = this.args;
        if (field.type instanceof GraphQLNonNull && field.type.ofType instanceof GraphQLScalarType) {
            field.type = new GraphQLNonNull(new CanVisitType(field.type.ofType));
        }
        else if (field.type instanceof GraphQLScalarType) {
            field.type = new CanVisitType(field.type);
        }
        else {
            throw new Error(`Not a scalar type: ${field.type}`);
        }
        field.resolve = function (source, args, context, info) {
            return __awaiter(this, void 0, void 0, function* () {
                const action = info.operation.operation;
                const path = Info2Path(info);
                let role = context[CasbinContextRoleKey];
                if (!(CasbinContextEnforcerKey in context)) {
                    throw new EnforcerNotFoundError(`You must provide Enforcer in the context with key ${CasbinContextEnforcerKey}`);
                }
                const CasbinInstance = context[CasbinContextEnforcerKey];
                if (!(CasbinContextPathCacheKey in context)) {
                    context[CasbinContextPathCacheKey] = yield buildPathCache(CasbinInstance);
                }
                const pathCache = context[CasbinContextPathCacheKey];
                if (!pathCache.includes(path)) {
                    yield CasbinInstance.addPolicy(who, path, action);
                    pathCache.push(PathAndAction2Key(path, action));
                }
                const passed = yield CasbinInstance.enforce(role, path, action);
                if (passed) {
                    return yield resolve.apply(this, [source, args, context, info]);
                }
                else {
                    if (role === '*') {
                        role = 'anonymous';
                    }
                    throw new PermissionInsufficientError(`${role} can not ${action} ${path}`);
                }
            });
        };
    }
}
/**
 * Override GraphQLScalarType, used to judge if current node have directive.
 */
export class CanVisitType extends GraphQLScalarType {
    constructor(type) {
        super({
            name: CasbinAuthTypeName,
            serialize(value) {
                return type.serialize(value);
            },
            parseLiteral(value, maybe) {
                return type.parseLiteral(value, maybe);
            },
            parseValue(value) {
                return type.parseValue(value);
            },
        });
    }
}
